package cn.estsh.i3plus.core.apiservice.serviceimpl.base;

import cn.estsh.i3plus.core.api.iservice.base.IDataSeparatorService;
import cn.estsh.i3plus.core.apiservice.configuration.CoreJdbcTemplate;
import cn.estsh.i3plus.core.apiservice.configuration.CoreJdbcTemplateConfig;
import cn.estsh.i3plus.core.apiservice.util.SeparatorDataUtil;
import cn.estsh.i3plus.platform.common.convert.ConvertBean;
import cn.estsh.i3plus.platform.common.util.CommonConstWords;
import cn.estsh.i3plus.pojo.base.bean.DdlPackBean;
import cn.estsh.i3plus.pojo.base.bean.ListPager;
import cn.estsh.i3plus.pojo.base.common.Pager;
import cn.estsh.i3plus.pojo.base.common.PagerHelper;
import cn.estsh.i3plus.pojo.base.jpa.dao.BaseRepository;
import cn.estsh.i3plus.pojo.base.tool.DdlPreparedPack;
import cn.estsh.i3plus.pojo.platform.bean.DataSeparatorMessage;
import cn.estsh.i3plus.pojo.platform.bean.DataSeparatorRule;
import cn.estsh.i3plus.pojo.platform.repository.DataSeparatorRepository;
import cn.estsh.impp.framework.boot.auth.AuthUtil;
import cn.estsh.impp.framework.boot.exception.ImppExceptionBuilder;
import cn.estsh.impp.framework.boot.util.ImppRedis;
import com.alibaba.druid.DbType;
import com.alibaba.druid.sql.SQLUtils;
import com.alibaba.druid.sql.ast.SQLStatement;
import com.alibaba.druid.sql.visitor.ExportParameterizedOutputVisitor;
import com.alibaba.druid.util.JdbcConstants;
import com.google.common.base.Strings;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.jdbc.core.namedparam.NamedParameterJdbcTemplate;
import org.springframework.stereotype.Service;

import javax.annotation.Resource;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * @Description :
 * @Reference :
 * @Author : Castle
 * @CreateDate : 2021/12/3 13:44
 * @Modify:
 **/
@Service
public class DataSeparatorServiceImpl implements IDataSeparatorService {

    @Resource(name = CommonConstWords.IMPP_REDIS_RES)
    private ImppRedis redisRes;

    @Resource
    private DataSeparatorRepository dataSeparatorRepository;

    @Resource
    private CoreJdbcTemplateConfig coreJdbcTemplateConfig;

    @Autowired
    private SeparatorDataUtil separatorDataUtil;

    @Override
    public void doSeparate(BaseRepository baseRepository, DataSeparatorMessage msg) {
        //获取数据
        Long id = msg.getId();
        Object bean = baseRepository.getById(id);

        //获取目的地的数据源
        CoreJdbcTemplate coreJdbcTemplate = coreJdbcTemplateConfig.getCoreJdbcTemplate(msg.getDestDataSourceId());
        //迁移数据
        try {
            execute(bean, msg,coreJdbcTemplate);
        } catch (Exception e) {
            throw ImppExceptionBuilder.newInstance().setErrorDetail(e.getMessage()+":separate data failed").build();
        }
        //删除原数据
        baseRepository.deleteById(id);
    }

    @Override
    public boolean addRule(DataSeparatorRule msg) {
        ConvertBean.saveOrUpdate(msg, AuthUtil.getSessionUser().getUserName());
        DataSeparatorRule rule = dataSeparatorRepository.insert(msg);
        String redisKey = CommonConstWords.RES_DATA_SEPARATOR + ":" + rule.getRefBeanName().toUpperCase();
        redisRes.putObject(redisKey, rule, -1);
        return true;
    }

    @Override
    public boolean deleteRule(Long id) {
        DataSeparatorRule rule = dataSeparatorRepository.getById(id);
        String redisKey = CommonConstWords.RES_DATA_SEPARATOR + rule.getRefBeanName().toUpperCase();
        dataSeparatorRepository.deleteWeaklyById(id, AuthUtil.getSessionUser().getUserName());
        redisRes.deleteKey(redisKey);
        return true;
    }

    @Override
    public boolean update(DataSeparatorRule msg) {
        ConvertBean.saveOrUpdate(msg, AuthUtil.getSessionUser().getUserName());
        dataSeparatorRepository.update(msg);
        String redisKey = CommonConstWords.RES_DATA_SEPARATOR + ":" + msg.getRefBeanName().toUpperCase();
        redisRes.putObject(redisKey, msg, -1);
        return true;
    }

    @Override
    public ListPager<DataSeparatorRule> query(DataSeparatorRule rule, Pager pager) {
        DdlPackBean ddlPackBean = DdlPackBean.getDdlPackBean(rule.getOrganizeCode());

        if (Strings.isNullOrEmpty(rule.getRefBeanName())) {
            DdlPreparedPack.getStringLeftLikerPack(rule.getRefBeanName(), "refBeanName", ddlPackBean);
        }

        if (Strings.isNullOrEmpty(rule.getRuleColumn())) {
            DdlPreparedPack.getStringLeftLikerPack(rule.getRuleColumn(), "ruleColumn", ddlPackBean);
        }
        pager = PagerHelper.getPager(pager, dataSeparatorRepository.findByHqlWhereCount(ddlPackBean));

        return new ListPager(dataSeparatorRepository.findByHqlPage(ddlPackBean, pager), pager);
    }

    @Override
    public void separatorRedisInit() {
        //启动加载所有的数据到redis中
        List<DataSeparatorRule> list = dataSeparatorRepository.list();
        list.forEach(item -> {
            String refBeanName = item.getRefBeanName().toUpperCase();
            String redisKey = CommonConstWords.RES_DATA_SEPARATOR + refBeanName;
            redisRes.putObject(redisKey, item, -1);
        });
    }


    private void execute(Object bean,DataSeparatorMessage msg,CoreJdbcTemplate coreJdbcTemplate) throws Exception {
        String insertSql = separatorDataUtil.packSql(bean, msg.getDestTableName());
        NamedParameterJdbcTemplate jdbcTemplate = coreJdbcTemplate.getNamedParameterJdbcTemplate();
        boolean isSqlServer = isMSSQLSERVER(jdbcTemplate);
        DbType dbType;
        if (isSqlServer) {
            dbType = JdbcConstants.SQL_SERVER;
        } else {
            dbType = JdbcConstants.MYSQL;
        }
        Map<String, Object> resultMap = parseSQL(insertSql, dbType);
        String preparedSql = resultMap.get("preparedSql").toString();
        @SuppressWarnings("unchecked")
        List<Object> parameters = (List<Object>) resultMap.get("parameters");
        jdbcTemplate.getJdbcTemplate().update(preparedSql,parameters.toArray());
    }


    /**
     * 解析sql为预编sql和参数
     *
     * @param sql
     * @param dbType
     * @return
     */
    private Map<String, Object> parseSQL(String sql, DbType dbType) {

        List<SQLStatement> sqlStatements = SQLUtils.parseStatements(sql, dbType);

        ExportParameterizedOutputVisitor visitor = new ExportParameterizedOutputVisitor();
        for (SQLStatement statement : sqlStatements) {
            statement.accept(visitor);
        }
        String preparedSql = visitor.getAppender().toString();
        List<Object> parameters = visitor.getParameters();
        HashMap<String, Object> resultMap = new HashMap<>();
        resultMap.put("preparedSql", preparedSql);
        resultMap.put("parameters", parameters);
        return resultMap;
    }

    /**
     * 判断数据库是不是 MSSQLSERVER
     *
     * @return
     * @throws Exception
     */
    private boolean isMSSQLSERVER(NamedParameterJdbcTemplate srcJdbcTemplate) throws Exception {
        return srcJdbcTemplate.getJdbcTemplate().getDataSource().getConnection().getMetaData().getURL().contains("sqlserver");
    }
}
